-- set name and namespace
addonname, PAR = ...

-- shortcut
local CPB = C_PetBattles

-- lazy, easy condition evaluation functions
local function Any(v, t) for i = 1, #t do if (v == t[i]) then return true end end return false end
local function NotAny(v, t) for i = 1, #t do if (v == t[i]) then return false end end return true end

-- Debugging (remove hyphens from the following line to activate)
--local DEBUGGING = true

-- multiple effect handling
local lastabilityID
--local lastturnIndex - possible future improvement
local lasteffectIndex

-- create event handling frame and register event(s)
local PAREventFrame = CreateFrame("Frame", "PAREventFrame")
PAREventFrame:RegisterEvent("ADDON_LOADED")
PAREventFrame:RegisterEvent("CHAT_MSG_PET_BATTLE_COMBAT_LOG")

-- quick function for users to reset their data
local function ResetPARData()
	PARData = {}
	PARData.PlayerData = {}
	PARData.PlayerData.Miss = 0
	PARData.PlayerData.Hit = 0
	PARData.PlayerData.Total = 0
	PARData.EnemyData = {}
	PARData.EnemyData.Miss = 0
	PARData.EnemyData.Hit = 0
	PARData.EnemyData.Total = 0
end

-- event handling
local function PAREventFrame_OnEvent(self, event, message, ...)
	if (event == "ADDON_LOADED") and (message == addonname) then
		
		-- create initial SavedVariables if not found
		if (not PARData) then
			PARData = {}
			PARData.PlayerData = {}
			PARData.PlayerData.Miss = 0
			PARData.PlayerData.Hit = 0
			PARData.PlayerData.Total = 0
			PARData.EnemyData = {}
			PARData.EnemyData.Miss = 0
			PARData.EnemyData.Hit = 0
			PARData.EnemyData.Total = 0
		end
		
		-- unregister event now that addon has loaded
		PAREventFrame:UnregisterEvent("ADDON_LOADED")
	elseif (event == "CHAT_MSG_PET_BATTLE_COMBAT_LOG") then
	
		-- RETURN if not a miss, buff/debuff application, or damaging attack
		if (not strmatch(message, " applied ")) and (not strmatch(message, " missed ")) and (not strmatch(message, " dealt ")) then
			if DEBUGGING then
				print("Not a miss, application, or damage.")
			end
			return
		end
		
		-- retrieve abilityID from message (first abilityID is only relevant one, so ignoring possiblity of others)
		local startpos = strfind(message, "battlePetAbil")
		local _, endpos = strfind(message, "|h|r")
		local _, abilityID, _ = strsplit(":", strsub(message, startpos, endpos))
		abilityID = tonumber(abilityID)
		
		-- RETURN if abilityID matches any RejectedAbilities
		if (not abilityID) or Any(abilityID, PAR.RejectedAbilities) then
			if DEBUGGING and abilityID then
				print("Ability with ID " .. abilityID .. " is not on the list of acceptable abilities.")
			end
			return
		end
		
		-- Find out which player is doing the attack
		-- " enemy " means you did the attack
		-- " your " means the enemy did the attack
		local petOwner, enemyOwner
		if strmatch(message, " enemy ") then
			petOwner = 1
			enemyOwner = 2
		elseif strmatch(message, " your ") then
			petOwner = 2
			enemyOwner = 1
		else
			if DEBUGGING then
				print("Couldn't figure out petOwner/enemyOwner.")
			end
			return
		end
		
		-- find active pet and RETURN if missing
		local petIndex = CPB.GetActivePet(petOwner)
		local enemyIndex = CPB.GetActivePet(enemyOwner)
		if not (petIndex and enemyIndex) then
			if DEBUGGING then
				print("Active pet could not be found.")
			end
			return
		end
		
		-- RETURN if pets are not the same level
		local petLevel = CPB.GetLevel(petOwner, petIndex)
		local enemyLevel = CPB.GetLevel(enemyOwner, enemyIndex)
		if (not (petLevel and enemyLevel)) or (petLevel ~= enemyLevel) then
			if DEBUGGING then
				print("Active pets are not the same level.")
			end
			return
		end
		
		-- RETURN if opposite pet has an aura from MissAuras
		local numAuras = CPB.GetNumAuras(enemyOwner, enemyIndex)
		if (numAuras) and (numAuras > 0) then
			for auraIndex = 1, numAuras do
				local currentAura = CPB.GetAuraInfo(enemyOwner, enemyIndex, auraIndex)
				if Any(currentAura, PAR.MissAuras) then
					if DEBUGGING then
						print("There is an Aura on the enemy pet that causes misses.")
					end
					return
				end
			end	
		end
		
		-- calculate AccuracyModifier using Blizzard's StateValues (alternate approach: use AccuracyAuras array)
		local AccuracyModifier
		if (CPB.GetPetType(petOwner, petIndex) == 7) then -- Elementals aren't affected by weathers.
			AccuracyModifier = CPB.GetStateValue(petOwner, petIndex, 41) + CPB.GetStateValue(petOwner, 0, 41)
		else
			AccuracyModifier = CPB.GetStateValue(petOwner, petIndex, 41) + CPB.GetStateValue(petOwner, 0, 41) + CPB.GetStateValue(0, 0, 41)
		end
		
		-- determine if ability was the last ability used and reset effect index if needed
		if (not lastabilityID) or (lastabilityID ~= abilityID) then
			lastabilityID = abilityID
			lasteffectIndex = 0
		end
		
		-- RETURN if last effect index is invalid
		if (not lasteffectIndex) or (lasteffectIndex > 9) then
			if DEBUGGING then
				print("Last effect index was out of bounds.")
			end
			return
		end
		
		-- Obtain raw accuracy value for ability by looping through possible turns/effects
		-- this code was heavily modified in r6 to account for abilities with multiple effects
		local RawAccuracy
		--for turnIndex = 1, 10 do
		for effectIndex = lasteffectIndex + 1, 10 do
			RawAccuracy = CPB.GetAbilityEffectInfo(abilityID, 1, effectIndex, "accuracy")
			if RawAccuracy then 
				lasteffectIndex = effectIndex
				break
			end
		end
			--if RawAccuracy then break end
		--end
		
		-- RETURN if we don't find any accuracy value
		if not RawAccuracy then
			if DEBUGGING then
				print("Error, found no accuracy result for ability with ID " .. abilityID .. " and last effect index " .. lasteffectIndex .. ".")
			end
			return
		end
		
		-- combine Accuracy and make sure it stays within values that make sense
		local totalAccuracy = RawAccuracy + AccuracyModifier
		if (totalAccuracy < 0) then totalAccuracy = 0 end
		if (totalAccuracy > 100) then totalAccuracy = 100 end
		
		-- record data, separate into miss and hit (dealt/applied)
		-- table structure is PARData[totalAccuracy][abilityID].Miss (and Hit and Total)
		-- then follow up by looking at PARData[totalAccuracy].Miss
		-- and lastly PARData.Miss
		if (petOwner == 1) then
			
			-- create table structure if empty previously
			if (not PARData.PlayerData[totalAccuracy]) then
				PARData.PlayerData[totalAccuracy] = {}
				PARData.PlayerData[totalAccuracy].Miss = 0
				PARData.PlayerData[totalAccuracy].Hit = 0
				PARData.PlayerData[totalAccuracy].Total = 0
			end
			if (not PARData.PlayerData[totalAccuracy][abilityID]) then
				PARData.PlayerData[totalAccuracy][abilityID] = {}
				PARData.PlayerData[totalAccuracy][abilityID].Miss = 0
				PARData.PlayerData[totalAccuracy][abilityID].Hit = 0
				PARData.PlayerData[totalAccuracy][abilityID].Total = 0
			end
			if (not PARData.PlayerData[totalAccuracy][abilityID][lasteffectIndex]) then
				PARData.PlayerData[totalAccuracy][abilityID][lasteffectIndex] = {}
				PARData.PlayerData[totalAccuracy][abilityID][lasteffectIndex].Miss = 0
				PARData.PlayerData[totalAccuracy][abilityID][lasteffectIndex].Hit = 0
				PARData.PlayerData[totalAccuracy][abilityID][lasteffectIndex].Total = 0
			end
			
			if (strmatch(message, " applied ")) or (strmatch(message, " dealt ")) then
				
				-- add a "Hit" result
				PARData.PlayerData[totalAccuracy][abilityID][lasteffectIndex].Hit = PARData.PlayerData[totalAccuracy][abilityID][lasteffectIndex].Hit + 1
				PARData.PlayerData[totalAccuracy][abilityID].Hit = PARData.PlayerData[totalAccuracy][abilityID].Hit + 1
				PARData.PlayerData[totalAccuracy].Hit = PARData.PlayerData[totalAccuracy].Hit + 1
				PARData.PlayerData.Hit = PARData.PlayerData.Hit + 1
				
				-- increment total number of recorded attacks
				PARData.PlayerData[totalAccuracy][abilityID][lasteffectIndex].Total = PARData.PlayerData[totalAccuracy][abilityID][lasteffectIndex].Total + 1
				PARData.PlayerData[totalAccuracy][abilityID].Total = PARData.PlayerData[totalAccuracy][abilityID].Total + 1
				PARData.PlayerData[totalAccuracy].Total = PARData.PlayerData[totalAccuracy].Total + 1
				PARData.PlayerData.Total = PARData.PlayerData.Total + 1
			elseif (strmatch(message, " missed ")) then
				
				-- add a "Miss" result
				PARData.PlayerData[totalAccuracy][abilityID][lasteffectIndex].Miss = PARData.PlayerData[totalAccuracy][abilityID][lasteffectIndex].Miss + 1
				PARData.PlayerData[totalAccuracy][abilityID].Miss = PARData.PlayerData[totalAccuracy][abilityID].Miss + 1
				PARData.PlayerData[totalAccuracy].Miss = PARData.PlayerData[totalAccuracy].Miss + 1
				PARData.PlayerData.Miss = PARData.PlayerData.Miss + 1
				
				-- increment total number of recorded attacks
				PARData.PlayerData[totalAccuracy][abilityID][lasteffectIndex].Total = PARData.PlayerData[totalAccuracy][abilityID][lasteffectIndex].Total + 1
				PARData.PlayerData[totalAccuracy][abilityID].Total = PARData.PlayerData[totalAccuracy][abilityID].Total + 1
				PARData.PlayerData[totalAccuracy].Total = PARData.PlayerData[totalAccuracy].Total + 1
				PARData.PlayerData.Total = PARData.PlayerData.Total + 1
			else
				if DEBUGGING then
					print("Bizarre error with string matching on first pet owner.")
				end
				return
			end
		elseif (petOwner == 2) then
		
			-- create table structure if empty previously
			if (not PARData.EnemyData[totalAccuracy]) then
				PARData.EnemyData[totalAccuracy] = {}
				PARData.EnemyData[totalAccuracy].Miss = 0
				PARData.EnemyData[totalAccuracy].Hit = 0
				PARData.EnemyData[totalAccuracy].Total = 0
			end
			if (not PARData.EnemyData[totalAccuracy][abilityID]) then
				PARData.EnemyData[totalAccuracy][abilityID] = {}
				PARData.EnemyData[totalAccuracy][abilityID].Miss = 0
				PARData.EnemyData[totalAccuracy][abilityID].Hit = 0
				PARData.EnemyData[totalAccuracy][abilityID].Total = 0
			end
			if (not PARData.EnemyData[totalAccuracy][abilityID][lasteffectIndex]) then
				PARData.EnemyData[totalAccuracy][abilityID][lasteffectIndex] = {}
				PARData.EnemyData[totalAccuracy][abilityID][lasteffectIndex].Miss = 0
				PARData.EnemyData[totalAccuracy][abilityID][lasteffectIndex].Hit = 0
				PARData.EnemyData[totalAccuracy][abilityID][lasteffectIndex].Total = 0
			end
			
			if (strmatch(message, " applied ")) or (strmatch(message, " dealt ")) then
				
				-- add a "Hit" result
				PARData.EnemyData[totalAccuracy][abilityID][lasteffectIndex].Hit = PARData.EnemyData[totalAccuracy][abilityID][lasteffectIndex].Hit + 1
				PARData.EnemyData[totalAccuracy][abilityID].Hit = PARData.EnemyData[totalAccuracy][abilityID].Hit + 1
				PARData.EnemyData[totalAccuracy].Hit = PARData.EnemyData[totalAccuracy].Hit + 1
				PARData.EnemyData.Hit = PARData.EnemyData.Hit + 1
				
				-- increment total number of recorded attacks
				PARData.EnemyData[totalAccuracy][abilityID][lasteffectIndex].Total = PARData.EnemyData[totalAccuracy][abilityID][lasteffectIndex].Total + 1
				PARData.EnemyData[totalAccuracy][abilityID].Total = PARData.EnemyData[totalAccuracy][abilityID].Total + 1
				PARData.EnemyData[totalAccuracy].Total = PARData.EnemyData[totalAccuracy].Total + 1
				PARData.EnemyData.Total = PARData.EnemyData.Total + 1
			elseif (strmatch(message, " missed ")) then
				
				-- add a "Miss" result
				PARData.EnemyData[totalAccuracy][abilityID][lasteffectIndex].Miss = PARData.EnemyData[totalAccuracy][abilityID][lasteffectIndex].Miss + 1
				PARData.EnemyData[totalAccuracy][abilityID].Miss = PARData.EnemyData[totalAccuracy][abilityID].Miss + 1
				PARData.EnemyData[totalAccuracy].Miss = PARData.EnemyData[totalAccuracy].Miss + 1
				PARData.EnemyData.Miss = PARData.EnemyData.Miss + 1
				
				-- increment total number of recorded attacks
				PARData.EnemyData[totalAccuracy][abilityID][lasteffectIndex].Total = PARData.EnemyData[totalAccuracy][abilityID][lasteffectIndex].Total + 1
				PARData.EnemyData[totalAccuracy][abilityID].Total = PARData.EnemyData[totalAccuracy][abilityID].Total + 1
				PARData.EnemyData[totalAccuracy].Total = PARData.EnemyData[totalAccuracy].Total + 1
				PARData.EnemyData.Total = PARData.EnemyData.Total + 1
			else
				if DEBUGGING then
					print("Bizarre error with string matching on second pet owner.")
				end
				return
			end
		else
			if DEBUGGING then
				print("Bizarre error with pet ownership.")
			end
			return
		end
	end
end

-- set our event handler function
PAREventFrame:SetScript("OnEvent", PAREventFrame_OnEvent)

-- slash command handling
SLASH_PETACCURACYRECORDER1 = "/PAR"
SLASH_PETACCURACYRECORDER2 = "/petaccuracy"
SlashCmdList["PETACCURACYRECORDER"] = function(message)
	
	-- lowercase message for evaluation
	local msg = string.lower(message)
	
	-- RETURN if no data or invalid data or user requested it
	if (msg == "reset") or (not PARData.PlayerData.Total) or (not PARData.EnemyData.Total) or (PARData.PlayerData.Total == 0) or (PARData.EnemyData.Total == 0) then
		if (msg == "reset") then print("Pet Accuracy Recorder: Resetting all recorded data!") end
		ResetPARData()
		return
	end
	
	--
	-- DO STATS FOR PLAYER DATA
	--
	
	-- first find ExpectedOverallAccuracy
	local ExpectedOverallAccuracy = 0
	
	-- use totalAccuracy * prevalence, where prevalence is calculated as (PARData[totalAccuracy].Total / PARData.Total)
	for k, v in next, PARData.PlayerData do
		if (k) and (type(v) == "table") then
			ExpectedOverallAccuracy = ExpectedOverallAccuracy + ((k / 100) * (PARData.PlayerData[k].Total / PARData.PlayerData.Total))
		end
	end
	
	-- next find TrueOverallAccuracy, calculated as (PARData.Hit / PARData.Total)
	local TrueOverallAccuracy = PARData.PlayerData.Hit / PARData.PlayerData.Total
	
	-- save results to data table
	PARData.PlayerData.Expected = ExpectedOverallAccuracy
	PARData.PlayerData.True = TrueOverallAccuracy
	
	-- show results to user
	print(format("Expected Accuracy for Player: %.2f%%", PARData.PlayerData.Expected * 100))
	print(format("Actual Accuracy for Player: %.2f%%", PARData.PlayerData.True * 100))
	print(format("Sample Size: %i attacks", PARData.PlayerData.Total))
	
	-- do statistical analysis (obtain std deviation and perform t-test) in future version
	--[[for k, v in next, PARData do
		if (k) and (type(v) == "table") then
			ExpectedOverallAccuracy = ExpectedOverallAccuracy + ((k / 100) * (PARData[k].Total / PARData.Total))
		end
	end]]--
	
	--
	-- DO STATS FOR ENEMY DATA
	--
	
	-- first find ExpectedOverallAccuracy
	local ExpectedOverallAccuracy = 0
	
	-- use totalAccuracy * prevalence, where prevalence is calculated as (PARData[totalAccuracy].Total / PARData.Total)
	for k, v in next, PARData.EnemyData do
		if (k) and (type(v) == "table") then
			ExpectedOverallAccuracy = ExpectedOverallAccuracy + ((k / 100) * (PARData.EnemyData[k].Total / PARData.EnemyData.Total))
		end
	end
	
	-- next find TrueOverallAccuracy, calculated as (PARData.Hit / PARData.Total)
	local TrueOverallAccuracy = PARData.EnemyData.Hit / PARData.EnemyData.Total
	
	-- save results to data table
	PARData.EnemyData.Expected = ExpectedOverallAccuracy
	PARData.EnemyData.True = TrueOverallAccuracy
	
	-- show results to user
	print(format("Expected Accuracy for Enemy: %.2f%%", PARData.EnemyData.Expected * 100))
	print(format("Actual Accuracy for Enemy: %.2f%%", PARData.EnemyData.True * 100))
	print(format("Sample Size: %i attacks", PARData.EnemyData.Total))
	
	-- do statistical analysis (obtain std deviation and perform t-test) in future version
	--[[for k, v in next, PARData do
		if (k) and (type(v) == "table") then
			ExpectedOverallAccuracy = ExpectedOverallAccuracy + ((k / 100) * (PARData[k].Total / PARData.Total))
		end
	end]]--
end